/*
 * Jack-detection / reporting
 *
 * Copyright (c) 2009 Takashi Iwai <tiwai@suse.de>
 *
 *  This driver is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This driver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <linux/slab.h>
#include <linux/pci.h>
#include <sound/core.h>
#include <sound/jack.h>
#include "hda_codec.h"
#include "hda_local.h"

struct hda_jack_item {
	struct hda_codec *codec;
	struct snd_kcontrol *ctl;
#ifdef CONFIG_SND_JACK
	struct snd_jack *jack;
#endif
	char name[32];
	hda_nid_t nid;
	int jack_type;
	unsigned int state;
	unsigned int flags;
	struct list_head list;
};

#ifdef CONFIG_SND_JACK
static void hda_jack_free_jack_priv(struct snd_jack *jack)
{
	struct hda_jack_item *item = jack->private_data;
	item->jack = NULL;
}
#endif

static void hda_jack_free_ctl_priv(struct snd_kcontrol *kcontrol)
{
	struct hda_jack_item *item = snd_kcontrol_chip(kcontrol);
	item->ctl = NULL;
}

static int hda_jack_item_update(struct hda_jack_item *item)
{
	unsigned int state;
	unsigned pincap = snd_hda_query_pin_caps(item->codec, item->nid);
	if (pincap & AC_PINCAP_TRIG_REQ)
		snd_hda_codec_read(item->codec, item->nid, 0,
				   AC_VERB_SET_PIN_SENSE, 0);
	state = snd_hda_codec_read(item->codec, item->nid, 0,
				   AC_VERB_GET_PIN_SENSE, 0);
	if (state == item->state)
		return 0;
	item->state = state;
	return 1;
}

static int hda_jack_is_plugged(struct hda_jack_item *item)
{
	if (item->flags & SND_HDA_JACK_SENSE_INVERSE)
		return !(item->state & AC_PINSENSE_PRESENCE);
	else
		return !!(item->state & AC_PINSENSE_PRESENCE);
}

#define hda_jack_ctl_info	snd_ctl_boolean_mono_info

static int hda_jack_ctl_get(struct snd_kcontrol *kcontrol,
			    struct snd_ctl_elem_value *ucontrol)
{
	struct hda_jack_item *item = snd_kcontrol_chip(kcontrol);

	if (!item->nid)
		return -EINVAL;
	hda_jack_item_update(item);
	ucontrol->value.integer.value[0] = hda_jack_is_plugged(item);
	return 0;
}

static struct snd_kcontrol_new kctl_temp = {
	.access = SNDRV_CTL_ELEM_ACCESS_READ,
	.iface = SNDRV_CTL_ELEM_IFACE_CARD, /* not a mixer iface! */
	/* name is filled later */
	.info = hda_jack_ctl_info,
	.get = hda_jack_ctl_get,
};

static void hda_jack_item_free(struct hda_jack_item *item)
{
	list_del(&item->list);
#ifdef CONFIG_SND_JACK
	if (item->jack)
		snd_device_free(item->codec->bus->card, item->jack);
#endif
	if (item->ctl)
		snd_ctl_free_one(item->ctl);
	kfree(item);
}

int __snd_hda_jack_add(struct hda_codec *codec, hda_nid_t nid, const char *name,
		       int jack_type, unsigned int flags)
{
	unsigned int def_conf = snd_hda_codec_get_pincfg(codec, nid);
	unsigned int connectivity = get_defcfg_connect(def_conf);
	unsigned int pincap;
	struct hda_jack_item *item;
	int err;

	if (connectivity == AC_JACK_PORT_NONE ||
	    connectivity == AC_JACK_PORT_FIXED) {
		snd_printdd("hda-codec: no valid connectivity (0x%x) "
			    "for jack 0x%x\n", connectivity, nid);
		return 0;
	}
	pincap = snd_hda_query_pin_caps(codec, nid);
	if (!(pincap & AC_PINCAP_PRES_DETECT)) {
		snd_printdd("hda-codec: jack 0x%x has no detect-capability\n",
			    nid);
		return 0;
	}

	item = kzalloc(sizeof(*item), GFP_KERNEL);
	if (!item)
		return -ENOMEM;
	INIT_LIST_HEAD(&item->list);
	item->codec = codec;
	item->nid = nid;
	item->jack_type = jack_type;
	item->flags = flags;

	if (name)
		strlcpy(item->name, name, sizeof(item->name));
	else {
		snprintf(item->name, sizeof(item->name),
			 "Jack %s at %s %s",
			 snd_hda_get_jack_type(def_conf),
			 snd_hda_get_jack_connectivity(def_conf),
			 snd_hda_get_jack_location(def_conf));
	}

#ifdef CONFIG_SND_JACK
	err = snd_jack_new(codec->bus->card, item->name, jack_type,
			   &item->jack);
	if (err < 0) {
		kfree(item);
		return err;
	}
	item->jack->private_data = item;
	item->jack->private_free = hda_jack_free_jack_priv;
#endif

	item->ctl = snd_ctl_new1(&kctl_temp, item);
	if (!item->ctl) {
		err = -ENOMEM;
		goto error;
	}
	strlcpy(item->ctl->id.name, item->name, sizeof(item->ctl->id.name));
	item->ctl->private_free = hda_jack_free_ctl_priv;

	err = snd_hda_ctl_add(codec, item->ctl);
	if (err < 0)
		goto error;

	list_add_tail(&item->list, &codec->jack_list_head);

	return 0;

 error:
	hda_jack_item_free(item);
	return err;
}
EXPORT_SYMBOL_HDA(__snd_hda_jack_add);

#ifdef CONFIG_SND_JACK
static void hda_report_jack(struct hda_codec *codec,
			    struct hda_jack_item *item)
{
	int type;
	if (hda_jack_is_plugged(item))
		type = item->jack_type;
	else
		type = 0;
	if (type == (SND_JACK_LINEOUT | SND_JACK_HEADPHONE)) {
		unsigned int pin_ctl;
		pin_ctl = snd_hda_codec_read(codec, item->nid, 0,
					     AC_VERB_GET_PIN_WIDGET_CONTROL,
					     0x00);
		type = (pin_ctl & AC_PINCTL_HP_EN)
			? SND_JACK_HEADPHONE : SND_JACK_LINEOUT;
	}
	snd_jack_report(item->jack, type);
}
#endif

int snd_hda_jack_report(struct hda_codec *codec, hda_nid_t nid)
{
	struct hda_jack_item *item;

	list_for_each_entry(item, &codec->jack_list_head, list) {
		if (item->nid != nid)
			continue;
		if (hda_jack_item_update(item))
			snd_ctl_notify(codec->bus->card,
				       SNDRV_CTL_EVENT_MASK_VALUE |
				       SNDRV_CTL_EVENT_MASK_INFO,
				       &item->ctl->id);
#ifdef CONFIG_SND_JACK
		hda_report_jack(codec, item);
#endif
		return 1;
	}
	return 0;
}
EXPORT_SYMBOL_HDA(snd_hda_jack_report);

void snd_hda_jack_free(struct hda_codec *codec)
{
	struct hda_jack_item *item;

	while (!list_empty(&codec->jack_list_head)) {
		item = list_first_entry(&codec->jack_list_head,
					struct hda_jack_item, list);
		hda_jack_item_free(item);
	}
}
